#include <algorithm>
#include <iostream>
#include <cerrno>
#include <cstring>
#include <stdint.h>
#include <sys/mman.h>

using namespace std;

// comments in upppercase refer to the guest code, followed
// by the corresponding host code in lowercase
char code[] = {
						// PROLOGUE
	0x50,                   		//   push   %eax
	0x51,                   		//   push   %ecx
						// PUSHL %EBP
	0x83, 0x6e, 0x10, 0x04,          	//   subl   $0x4,0x10(%esi)
	0x8b, 0x46, 0x14,             		//   mov    0x14(%esi),%eax
	0x8b, 0x4e, 0x10,             		//   mov    0x10(%esi),%ecx
	0x89, 0x04, 0x0f,             		//   mov    %eax,(%edi,%ecx,1)
						// MOVL %ESP, %EBP
	0x8b, 0x46, 0x10,             		//   mov    0x10(%esi),%eax
	0x89, 0x46, 0x14,             		//   mov    %eax,0x14(%esi)
						// MOVL 12(%EBP), %EAX
	0x8b, 0x4e, 0x14,             		//   mov    0x14(%esi),%ecx
	0x83, 0xc1, 0x0c,             		//   add    $0xc,%ecx
	0x8b, 0x04, 0x0f,             		//   mov    (%edi,%ecx,1),%eax
	0x89, 0x06,                		//   mov    %eax,(%esi)
						// MOVL 8(%EBP), %EDX
	0x8b, 0x4e, 0x14,             		//   mov    0x14(%esi),%ecx
	0x83, 0xc1, 0x08,             		//   add    $0x8,%ecx
	0x8b, 0x04, 0x0f,             		//   mov    (%edi,%ecx,1),%eax
	0x89, 0x46, 0x08,             		//   mov    %eax,0x8(%esi)
						// ADDL %EDX, %EAX
	0x8b, 0x46, 0x08,             		//   mov    0x8(%esi),%eax
	0x01, 0x06,                		//   add    %eax,(%esi)
	0x9c,                   		//   pushf
	0x8f, 0x46, 0x24,             		//   popl   0x24(%esi)
						// LEAVE
	0x8b, 0x46, 0x14,             		//   mov    0x14(%esi),%eax
	0x89, 0x46, 0x10,             		//   mov    %eax,0x10(%esi)
	0x8b, 0x04, 0x06,             		//   mov    (%esi,%eax,1),%eax
	0x89, 0x46, 0x14,             		//   mov    %eax,0x14(%esi)
	0x83, 0x46, 0x10, 0x04,          	//   addl   $0x4,0x10(%esi)
						// RET
	0x8b, 0x4e, 0x10,             		//   mov    0x10(%esi),%ecx
	0x8b, 0x04, 0x0f,             		//   mov    (%edi,%ecx,1),%eax
	0x89, 0x46, 0x20,             		//   mov    %eax,0x20(%esi)
	0x83, 0x46, 0x10, 0x04,          	//   addl   $0x4,0x10(%esi)
						// EPILOGUE
	0x59,                   		//   pop    %ecx
	0x58,                   		//   pop    %eax
	0xc3                    		//   ret
};

// the data structure that implements the guest CPU
struct CPU_des {
	uint32_t 	EAX;
	uint32_t	ECX;
	uint32_t	EDX;
	uint32_t	EBX;
	uint32_t	ESP;
	uint32_t	EBP;
	uint32_t	ESI;
	uint32_t	EDI;
	uint32_t	EIP;
	uint32_t	EFLAGS;
} CPU;

ostream& operator<<(ostream& o, const CPU_des& c)
{
	o << hex;
	o << "EAX\t" 	<< c.EAX 	<< endl;
	o << "ECX\t" 	<< c.ECX 	<< endl;
	o << "EDX\t" 	<< c.EDX 	<< endl;
	o << "EBX\t" 	<< c.EBX 	<< endl;
	o << "ESP\t" 	<< c.ESP 	<< endl;
	o << "EBP\t" 	<< c.EBP 	<< endl;
	o << "ESI\t" 	<< c.ESI 	<< endl;
	o << "EDI\t" 	<< c.EDI 	<< endl;
	o << "EIP\t" 	<< c.EIP 	<< endl;
	o << "EFLAGS\t" << c.EFLAGS;
	return o;
}

// guest memory
const int MEM_SIZE = 65536; // 64 KiB

uint8_t Mem[MEM_SIZE];


// the guest environment, containing pointers to the
// guest CPU and guest memory
struct env_des {
	CPU_des *CPU;
	uint8_t *MEM;
};

// exec() is defined in exec.s
extern "C" void exec(void *code, env_des* des);
int main()
{
	/* get some executable memory */
	void *exec_area = mmap(
		NULL,				// preferred addres: let the
						// kernel choose
		4096,				// requested size (a page)
		PROT_WRITE|PROT_READ|PROT_EXEC,	// protection bits: rwx
		MAP_PRIVATE|MAP_ANON,		// not backed by file, do not
						// share with other processes
		-1, 0				// file descriptor and offset:
						// unused with MAP_ANONYMOUS
	);
	if (exec_area == MAP_FAILED) {
		cerr << strerror(errno) << endl;
		return 1;
	}

	/* copy the code in the executable memory area */
	copy(code, code + sizeof(code), static_cast<char*>(exec_area));

	/* we prepare the CPU/Mem state to emulate a call to sum(2, 3) */

	// 
	CPU.ESP = MEM_SIZE;
	// PUSHL $3
	CPU.ESP -= 4;
	*reinterpret_cast<uint32_t *>(&Mem[CPU.ESP]) = 3;
	// PUSHL $2
	CPU.ESP -= 4;
	*reinterpret_cast<uint32_t *>(&Mem[CPU.ESP]) = 2;
	// CALL sum
	// we should push the return address, but we have
	// nowhere to return, so we simply decrement the
	// stack pointer
	CPU.ESP -= 4;
	// we should also set CPU.EIP, but we omit it in this
	// example, since we have just one translation block to
	// execute

	// initialize the environment
	env_des env = { &CPU, Mem };

	// call exec to jump to the translated code
	exec(exec_area, &env);

	// CPU.EAX should contain 5
	cout << "CPU state after execution: " << endl;
	cout << CPU << endl;

	return 0;
}
